/**********************************************************************/
/*   Color  terminal  emulator using the ctw (Color terminal widget)  */
/*   widget. This module is based on the athena widgets.	      */
/*   								      */
/*   Author: Paul Fox						      */
/*   Date:   January 1992       				      */
/**********************************************************************/
# include	"fcterm.h"
# include	<X11/Xaw/Paned.h>
# include	<X11/Xaw/Scrollbar.h>
# include	<X11/Xaw/Command.h>
# include	<X11/Xaw/MenuButton.h>
# include	<X11/Xaw/SimpleMenu.h>
# include	<X11/Xaw/SmeBSB.h>
# include	<X11/Xaw/SmeLine.h>
# include	<X11/Xaw/Toggle.h>
# include	<X11/Xaw/Command.h>

static Widget		font_dialog;
static Widget		dialog;
static Widget		monitor;
Widget		pshell;
Widget		top_level;
Widget		pane;
Widget		scroll_bar;
Widget		ctw_widget;
Widget		w_attr[CTW_MAX_ATTR];
Widget		w_status[4];
Widget		w_fonts[4];

XtIntervalId	timer_id;

char	*font_names[6];

XtAppContext	app_con;
int	pid;

/**********************************************************************/
/*   Copy of attributes which user can change.			      */
/**********************************************************************/
int	attributes[64];
char	**attr_names;

char	*prog_name;

static String fallback_resources[] = {
	"*foreground:		yellow",
	"*background:		black",
	"*menu*borderColor:	cyan",
	"*menu*borderWidth:	4",
	"*VT100.background: 	black",
	"*scrollbar*foreground:	magenta",
	"*allowShellResize:	true",
	(char *) NULL
	};

Widget	create_monitor_box();
void	jump_callback();
void	scroll_callback();
void	appl_callback();
void	input_callback();
void	top_callback();
void	menu_callback();
void	mouse_callback();
void	ok_dialog();
void	apply_dialog();
void	cancel_dialog();
void	refresh_dialog();
void	toggle_callback();
void	ok_status();
void	clear_status();
void	timeout_proc();
void	ok_font();
void	cancel_font();

int
main(argc, argv)
int	argc;
char	**argv;
{	int	n;
	Arg	args[20];
	Widget	menu_item;

	setup_signal_handlers();

	argc = steal_switches(argc, argv);

	n = 0;
	XtSetArg(args[n], XtNinput, TRUE); n++;
	XtSetArg(args[n], XtNallowShellResize, TRUE); n++;
	top_level = XtAppInitialize(&app_con, "XTerm", NULL, 0,
		&argc, argv, fallback_resources, args, n);


	do_switches(argc, argv);

	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane = XtCreateWidget(widget_name, panedWidgetClass,
			top_level, args, n);
	XtManageChild(pane);
	/***********************************************/
	/*   Create sub-shell.			       */
	/***********************************************/
	spawn_child();
	XtAppAddInput(app_con, pty_fd, XtInputReadMask, pty_input, NULL);

	/***********************************************/
	/*   Create scrollbar.			       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNwidth, 13); n++;
	scroll_bar = XtCreateWidget("scrollbar", scrollbarWidgetClass,
			pane, args, n);
	XawPanedSetMinMax(scroll_bar, 13, 13);
	XtAddCallback(scroll_bar, XtNjumpProc, jump_callback, "");
	XtAddCallback(scroll_bar, XtNscrollProc, scroll_callback, "");

	/***********************************************/
	/*   Create VT100 widget.		       */
	/***********************************************/
	n = 0;
	ctw_widget = XtCreateManagedWidget("VT100",
		ctwWidgetClass,
		pane, args, n);
	if (geometry) {
		int	w, h;
		int	fheight, fwidth;
		sscanf(geometry, "%dx%d", &w, &h);
		n = 0;
		XtSetArg(args[n], XtNcolumns, w); n++;
		XtSetArg(args[n], XtNrows, h); n++;
		get_font_size(ctw_widget, &fwidth, &fheight);
		XtSetArg(args[n], XtNwidth, fwidth * w); n++;
		XtSetArg(args[n], XtNheight, fheight * h); n++;
		XtSetValues(ctw_widget, args, n);

		n = 0;
		XtSetArg(args[n], XtNwidth, fwidth * w); n++;
		XtSetArg(args[n], XtNheight, fheight * h); n++;
		XtSetValues(scroll_bar, args, n);
		}
	XtAddCallback(ctw_widget, XtNapplCallback, appl_callback, 0);
	XtAddCallback(ctw_widget, XtNkbdCallback, input_callback, 0);
	XtAddCallback(ctw_widget, XtNresizeCallback, resize_callback, 0);
	XtAddCallback(ctw_widget, XtNtopCallback, top_callback, 0);
	XtAddCallback(ctw_widget, XtNmouseCallback, mouse_callback, 0);
	XawPanedAllowResize(ctw_widget, TRUE);

	XtRealizeWidget(top_level);

	create_icon_pixmap();

	set_default_attributes();

	/***********************************************/
	/*   Create popup menu.			       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNlabel, "Commands Menu Popup"); n++;
	pshell = XtCreatePopupShell("menu",
		simpleMenuWidgetClass,
		top_level,
		args, n);
	XtCreateManagedWidget("line", smeLineObjectClass, pshell, NULL, 0);
	menu_item = XtCreateManagedWidget(" Toggle Scrollbar",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, TOGGLE_SCROLLBAR);
	menu_item = XtCreateManagedWidget(" Reset Terminal",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, RESET_CHARACTERS);
	menu_item = XtCreateManagedWidget(" Paste",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, PASTE);
	menu_item = XtCreateManagedWidget(" Terminal attributes...",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, DIALOG_BOX);
	menu_item = XtCreateManagedWidget(" Status Monitor...",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, MONITOR);
	menu_item = XtCreateManagedWidget(" Change Font...",
		smeBSBObjectClass, pshell, NULL, 0);
	XtAddCallback(menu_item, XtNcallback, menu_callback, SET_FONT);

	stats.start_time = time((time_t *) NULL);
	XtAppMainLoop(app_con);
	exit(0);
	return 0;
}
/**********************************************************************/
/*   Function  to  handle  indication from VT100 widget that the top  */
/*   line  of  the window has moved so we can associate the thumb on  */
/*   the scrollbar with the position.				      */
/**********************************************************************/
void
top_callback(w, type, reason)
Widget	w;
int	type;
ctw_callback_t	*reason;
{	float	pos;
	int	n;
	Arg	args[20];
	int	save_lines;

	n = 0;
	XtSetArg(args[n], XtNsaveLines, &save_lines); n++;
	XtGetValues(w, args, n);
	pos = (float) reason->top_line / (float) save_lines; 
	if (pos >= (float) 0.99)
		pos = 0.99;
	XawScrollbarSetThumb(scroll_bar, pos, (float) 0.04);
}
/**********************************************************************/
/*   Handle a jump in the vertical scrollbar.			      */
/**********************************************************************/
void
jump_callback(widget, name, call_data)
Widget	widget;
char	*name;
caddr_t	call_data;
{	int	n;
	Arg	args[20];
	float	top = *(float *) call_data;
	int	save_lines;
	
	n = 0;
	XtSetArg(args[n], XtNsaveLines, &save_lines); n++;
	XtGetValues(ctw_widget, args, n);

	ctw_set_top_line(ctw_widget, (int) (save_lines * top));
}
/**********************************************************************/
/*   Handle a user moving the slider in the vertical scrollbar.	      */
/**********************************************************************/
void
scroll_callback(widget, name, call_data)
Widget	widget;
char	*name;
caddr_t	call_data;
{	int	n;
	Arg	args[20];
	int	scroll_fwd = (int) call_data > 0 ? TRUE : FALSE;
	int	num_rows;
	int	top_line;

	n = 0;
	XtSetArg(args[n], XtNrows, &num_rows); n++;
	XtGetValues(ctw_widget, args, n);
	top_line = ctw_get_top_line(ctw_widget);

	if (scroll_fwd)
		ctw_set_top_line(ctw_widget, top_line + num_rows);
	else
		ctw_set_top_line(ctw_widget, top_line - num_rows);
}
/**********************************************************************/
/*   Calback when menu item selected.				      */
/**********************************************************************/
void
mouse_callback(widget, name, reason)
Widget	widget;
char	*name;
ctw_callback_t	*reason;
{	int	n;
	Arg	args[20];
	Dimension	x, y;

	if (reason->reason != CTW_BUTTON_DOWN)
		return;

	n = 0;
	XtSetArg(args[n], XtNx, &x); n++;
	XtSetArg(args[n], XtNy, &y); n++;
	XtGetValues(top_level, args, n);

	n = 0;
	XtSetArg(args[n], XtNx, x + reason->event->xbutton.x - 5); n++;
	XtSetArg(args[n], XtNy, y + reason->event->xbutton.y - 5); n++;
	XtSetValues(pshell, args, n);
	if (!XtIsRealized(pshell))
		XtRealizeWidget(pshell);
	XtPopupSpringLoaded(pshell);
	/***********************************************/
	/*   Tell  Xt  that  we want all events to be  */
	/*   mapped  to  the  popup  so  it  looks  &  */
	/*   feels the way it should do.	       */
	/***********************************************/
  	XtGrabPointer(pshell, True,
		EnterWindowMask | LeaveWindowMask | ButtonReleaseMask,
		GrabModeAsync, GrabModeAsync, None, None,
		reason->event->xbutton.time);
}
/**********************************************************************/
/*   Calback when menu item selected.				      */
/**********************************************************************/
void
menu_callback(widget, val, call_data)
Widget	widget;
enum menu_items val;
caddr_t	call_data;
{	int	*flags;
	Widget	shell;
	int	i;
	int	rows, cols, r, c;
	int	n, len;
	Arg	args[20];
	char	buf[BUFSIZ];
	Widget	pane0, pane1, pane2;
	Widget	w1;

	switch (val) {
	  case TOGGLE_SCROLLBAR:
	  	if (XtIsManaged(scroll_bar))
			XtUnmanageChild(scroll_bar);
		else {
			XtUnmanageChild(ctw_widget);
			XtManageChild(scroll_bar);
			XtManageChild(ctw_widget);
			}
		break;
	  case RESET_CHARACTERS:
	  	ctw_add_string(ctw_widget, "\033(B\033[2J", 7);
	  	break;
	  case PASTE:
	  	ctw_get_selection(ctw_widget);
	  	break;
	  case DIALOG_BOX:
	  	if (dialog != (Widget) NULL) {
			XRaiseWindow(XtDisplay(dialog), XtWindow(dialog));
			break;
			}
	  	len = ctw_get_attributes(ctw_widget, &flags, &attr_names);
		memcpy((char *) attributes, (char *) flags, 
			len * sizeof (int));
		shell = XtCreatePopupShell("shell",
			transientShellWidgetClass, top_level, NULL, 0);
		dialog = shell;
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane0 = XtCreateManagedWidget("pane0", panedWidgetClass,
			shell, args, n);
		/***********************************************/
		/*   Create title of dialog box.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtCreateManagedWidget("Terminal Attributes",
			labelWidgetClass, pane0, args, n);
		/***********************************************/
		/*   Create row/column of attributes.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane2 = XtCreateManagedWidget("pane2", panedWidgetClass,
			pane0, args, n);
		cols = 3;
		rows = (len + cols - 1) / cols; 
		i = 0;
		for (c = 0; c < cols; c++) {
			n = 0;
			XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
			XtSetArg(args[n], XtNshowGrip, FALSE); n++;
			pane1 = XtCreateManagedWidget("pane3", panedWidgetClass,
				pane2, args, n);
			for (r = 0; r < rows; r++) {
				n = 0;
				switch (i) {
				  case CTW_SIZE:
				  case CTW_FONT_SIZE:
					sprintf(buf, "%s: %dx%d", attr_names[i], 
						flags[i] >> 16, flags[i] & 0xffff);
					break;
				  default:
					if (attr_names[i] == (char *) NULL) {
						buf[0] = '\0';
						XtSetArg(args[n], XtNsensitive, FALSE); n++;
						}
					else
						sprintf(buf, "%s: %s  ", attr_names[i], flags[i] ? "Yes" : "No");
					break;
				  }
				XtSetArg(args[n], XtNshowGrip, FALSE); n++;
				XtSetArg(args[n], XtNjustify, XtJustifyRight); n++;
				XtSetArg(args[n], XtNlabel, buf); n++;
				w1 = XtCreateManagedWidget(attr_names[i],
					toggleWidgetClass, pane1, args, n);
				XtAddCallback(w1, XtNcallback, toggle_callback, i);
				w_attr[i] = w1;
				if (attr_names[i])
					i++;
				}
			}
		/***********************************************/
		/*   Create OK/Cancel area.		       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
		pane1 = XtCreateManagedWidget("pane4",
			panedWidgetClass, pane0, args, n);
		/***********************************************/
		/*   Create blank space at left of buttons.    */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNlabel, ""); n++;
		w1 = XtCreateManagedWidget("", labelWidgetClass, pane1, args, n);
		XawPanedSetMinMax(w1, 140, 70);
		/***********************************************/
		/*   Create push buttons.		       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("  OK  ",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, ok_dialog, 0);
		XawPanedSetMinMax(w1, 70, 70);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("CANCEL",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, cancel_dialog, 0);
		XawPanedSetMinMax(w1, 70, 70);

		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("",
			labelWidgetClass, pane1, args, n);
		XawPanedSetMinMax(w1, 70, 70);

		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("APPLY",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, apply_dialog, 0);
		XawPanedSetMinMax(w1, 70, 70);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("REFRESH",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, refresh_dialog, 0);
		XawPanedSetMinMax(w1, 70, 70);

		XtPopup(shell, XtGrabNone);
	  	break;
	  case MONITOR:
	  	shell = create_monitor_box();
		if (shell) {
			XtPopup(shell, XtGrabNone);
			/***********************************************/
			/*   Create timer to keep updating display.    */
			/***********************************************/
			timer_id = XtAppAddTimeOut(app_con, 1000L, timeout_proc, NULL);
			}
	  	break;
	  case SET_FONT:
	  	if (font_dialog != (Widget) NULL) {
			XRaiseWindow(XtDisplay(font_dialog), XtWindow(font_dialog));
			break;
			}
		shell = XtCreatePopupShell("shell",
			transientShellWidgetClass, top_level, NULL, 0);
		/***********************************************/
		/*   Get list of fonts.			       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNfont1, &font_names[0]); n++;
		XtSetArg(args[n], XtNfont2, &font_names[1]); n++;
		XtSetArg(args[n], XtNfont3, &font_names[2]); n++;
		XtSetArg(args[n], XtNfont4, &font_names[3]); n++;
		XtSetArg(args[n], XtNfont5, &font_names[4]); n++;
		XtSetArg(args[n], XtNfont6, &font_names[5]); n++;
		XtGetValues(ctw_widget, args, n);

		font_dialog = shell;
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane0 = XtCreateManagedWidget("pane0", panedWidgetClass,
			shell, args, n);
		/***********************************************/
		/*   Create title of dialog box.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtCreateManagedWidget("fontMenu",
			labelWidgetClass, pane0, args, n);
		/***********************************************/
		/*   Create row/column of attributes.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane2 = XtCreateManagedWidget("pane2", panedWidgetClass,
			pane0, args, n);
		/***********************************************/
		/*   Create font1 & font2 toggle widgets.      */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane1 = XtCreateManagedWidget("pane3", panedWidgetClass,
			pane2, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNlabel, font_names[0]); n++;
		XtSetArg(args[n], XtNradioData, font_names[0]); n++;
		w_fonts[0] = XtCreateManagedWidget("font1", toggleWidgetClass,
			pane1, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNradioGroup, w_fonts[0]); n++;
		XtSetArg(args[n], XtNlabel, font_names[1]); n++;
		XtSetArg(args[n], XtNradioData, font_names[1]); n++;
		w_fonts[1] = XtCreateManagedWidget("font2", toggleWidgetClass,
			pane1, args, n);
		/***********************************************/
		/*   Create fonts 3 & 4			       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane1 = XtCreateManagedWidget("pane3", panedWidgetClass,
			pane2, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNradioGroup, w_fonts[0]); n++;
		XtSetArg(args[n], XtNlabel, font_names[2]); n++;
		XtSetArg(args[n], XtNradioData, font_names[2]); n++;
		w_fonts[2] = XtCreateManagedWidget("font3", toggleWidgetClass,
			pane1, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNradioGroup, w_fonts[0]); n++;
		XtSetArg(args[n], XtNlabel, font_names[3]); n++;
		XtSetArg(args[n], XtNradioData, font_names[3]); n++;
		w_fonts[3] = XtCreateManagedWidget("font4", toggleWidgetClass,
			pane1, args, n);
		/***********************************************/
		/*   Create fonts 5 & 6			       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		pane1 = XtCreateManagedWidget("pane4", panedWidgetClass,
			pane2, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNradioGroup, w_fonts[0]); n++;
		XtSetArg(args[n], XtNlabel, font_names[4]); n++;
		XtSetArg(args[n], XtNradioData, font_names[4]); n++;
		w_fonts[4] = XtCreateManagedWidget("font5", toggleWidgetClass,
			pane1, args, n);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNradioGroup, w_fonts[0]); n++;
		XtSetArg(args[n], XtNlabel, font_names[5]); n++;
		XtSetArg(args[n], XtNradioData, font_names[5]); n++;
		w_fonts[5] = XtCreateManagedWidget("font6", toggleWidgetClass,
			pane1, args, n);

		/***********************************************/
		/*   Create Button area.		       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
		pane1 = XtCreateManagedWidget("pane4",
			panedWidgetClass, pane0, args, n);
		/***********************************************/
		/*   Create blank space at left of buttons.    */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		XtSetArg(args[n], XtNlabel, ""); n++;
		w1 = XtCreateManagedWidget("", labelWidgetClass, pane1, args, n);
		XawPanedSetMinMax(w1, 140, 70);
		/***********************************************/
		/*   Create push buttons.		       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("  OK  ",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, ok_font, 0);
		XawPanedSetMinMax(w1, 70, 70);
		n = 0;
		XtSetArg(args[n], XtNshowGrip, FALSE); n++;
		w1 = XtCreateManagedWidget("CANCEL",
			commandWidgetClass, pane1, args, n);
		XtAddCallback(w1, XtNcallback, cancel_font, 0);
		XawPanedSetMinMax(w1, 70, 70);

		XtPopup(shell, XtGrabNone);
	  	break;
	  }
}
/**********************************************************************/
/*   Function to create the status monitor dialog box.		      */
/**********************************************************************/
Widget
create_monitor_box()
{	int	n;
	Arg	args[20];
	Widget	pane0, pane1, pane2;
	Widget	w1;
	char	buf[BUFSIZ];
	static Widget	shell;

	if (monitor != (Widget) NULL) {
		if (XtIsManaged(monitor)) {
			XRaiseWindow(XtDisplay(monitor), XtWindow(monitor));
			return NULL;
			}
		XtManageChild(monitor);
		return shell;
		}
	shell = XtCreatePopupShell("shell",
		transientShellWidgetClass, top_level, NULL, 0);
	monitor = shell;
	stats.time = time((time_t *) NULL);
	stats.bytes = 0;
	old_stats = stats;
	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane0 = XtCreateManagedWidget("pane0", panedWidgetClass,
		shell, args, n);
	/***********************************************/
	/*   Create title of dialog box.	       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtCreateManagedWidget("Status Monitor",
		labelWidgetClass, pane0, args, n);
	/***********************************************/
	/*   Create row/column of attributes.	       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane2 = XtCreateManagedWidget("pane2", panedWidgetClass,
		pane0, args, n);
	/***********************************************/
	/*   Total and cum byte count.		       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane1 = XtCreateManagedWidget("pane3", panedWidgetClass,
		pane2, args, n);
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	sprintf(buf, "Total bytes: %7ld ", stats.tot_bytes);
	XtSetArg(args[n], XtNlabel, buf); n++;
	w_status[0] = XtCreateManagedWidget("Total", labelWidgetClass,
		pane1, args, n);
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	sprintf(buf, "Cum bytes: %7ld ", stats.bytes);
	XtSetArg(args[n], XtNlabel, buf); n++;
	w_status[1] = XtCreateManagedWidget("Total", labelWidgetClass,
		pane1, args, n);
	/***********************************************/
	/*   Create the per/second bytes.	       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNorientation, XtorientVertical); n++;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	pane1 = XtCreateManagedWidget("pane3", panedWidgetClass,
		pane2, args, n);
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNlabel, "Total/sec:        "); n++;
	w_status[2] = XtCreateManagedWidget("Total", labelWidgetClass,
		pane1, args, n);
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNlabel, "Cum/sec:        "); n++;
	w_status[3] = XtCreateManagedWidget("Total", labelWidgetClass,
		pane1, args, n);

	/***********************************************/
	/*   Create Button area.		       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNorientation, XtorientHorizontal); n++;
	pane1 = XtCreateManagedWidget("pane4",
		panedWidgetClass, pane0, args, n);
	/***********************************************/
	/*   Create blank space at left of buttons.    */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	XtSetArg(args[n], XtNlabel, ""); n++;
	w1 = XtCreateManagedWidget("", labelWidgetClass, pane1, args, n);
	XawPanedSetMinMax(w1, 140, 70);
	/***********************************************/
	/*   Create push buttons.		       */
	/***********************************************/
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	w1 = XtCreateManagedWidget("  OK  ",
		commandWidgetClass, pane1, args, n);
	XtAddCallback(w1, XtNcallback, ok_status, 0);
	XawPanedSetMinMax(w1, 70, 70);
	n = 0;
	XtSetArg(args[n], XtNshowGrip, FALSE); n++;
	w1 = XtCreateManagedWidget("CLEAR",
		commandWidgetClass, pane1, args, n);
	XtAddCallback(w1, XtNcallback, clear_status, 0);
	XawPanedSetMinMax(w1, 70, 70);

	return shell;
}
/**********************************************************************/
/*   Function called when OK button selected in dialog box.	      */
/**********************************************************************/
void
ok_dialog(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	
	ctw_set_attributes(ctw_widget, attributes);
	XtDestroyWidget(dialog);
	dialog = (Widget) NULL;
}
/**********************************************************************/
/*   Function called to apply current attribute settings.	      */
/**********************************************************************/
void
apply_dialog(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	
	ctw_set_attributes(ctw_widget, attributes);
}
/**********************************************************************/
/*   Function  called  to  update  the  attributes  with the current  */
/*   settings.							      */
/**********************************************************************/
void
refresh_dialog(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{	int	*flags;
	int	len;
	int	i, n;
	Arg	args[20];
	char	buf[BUFSIZ];

  	len = ctw_get_attributes(ctw_widget, &flags, &attr_names);
	memcpy((char *) attributes, (char *) flags, len * sizeof (int));
	for (i = 0; i < len; i++) {
		switch (i) {
		  case CTW_SIZE:
		  case CTW_FONT_SIZE:
			sprintf(buf, "%s: %dx%d", attr_names[i], 
				flags[i] >> 16, flags[i] & 0xffff);
			break;
		  default:
			sprintf(buf, "%s: %s  ", attr_names[i], flags[i] ? "Yes" : "No");
			break;
		  }
		n = 0;
		XtSetArg(args[n], XtNlabel, buf); n++;
		XtSetValues(w_attr[i], args, n);
		}
}
/**********************************************************************/
/*   Function called when CANCEL button selected in dialog box.	      */
/**********************************************************************/
void
cancel_dialog(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	XtDestroyWidget(dialog);
	dialog = (Widget) NULL;
}
/**********************************************************************/
/*   Function  called  when  user toggles an attribute in the dialog  */
/*   box.							      */
/**********************************************************************/
void
toggle_callback(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{	int	i = (int) client_data;
	int	n;
	Arg	args[20];
	char	buf[BUFSIZ];

	if (i != CTW_SIZE && i != CTW_FONT_SIZE) {
		attributes[i] = !attributes[i];
		sprintf(buf, "%s: %s  ",
			attr_names[i], attributes[i] ? "Yes" : "No");
		n = 0;
		XtSetArg(args[n], XtNlabel, buf); n++;
		XtSetArg(args[n], XtNstate, FALSE); n++;
		XtSetValues(w, args, n);
		}
}
/**********************************************************************/
/*   Function to close down status monitor.			      */
/**********************************************************************/
void
ok_status(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	XtRemoveTimeOut(timer_id);
	XtUnmanageChild(monitor);
}
/**********************************************************************/
/*   Function to reset status					      */
/**********************************************************************/
void
clear_status(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	stats.time = time((time_t *) NULL);
	stats.bytes = 0;
}
/**********************************************************************/
/*   Timer function to update status display.			      */
/**********************************************************************/
void
timeout_proc(client_data, timer)
XtPointer	client_data;
XtIntervalId	*timer;
{	int	n;
	Arg	args[20];
	char	buf[BUFSIZ];
	time_t	t = time((time_t *) NULL);
	time_t	t1;

	if (stats.tot_bytes == old_stats.tot_bytes &&
	    stats.bytes == old_stats.bytes)
	    	goto restart_timer;

	if (stats.tot_bytes != old_stats.tot_bytes) {
		n = 0;
		sprintf(buf, "Total bytes: %ld", stats.tot_bytes);
		XtSetArg(args[n], XtNlabel, buf); n++;
		XtSetValues(w_status[0], args, n);
		}

	if (stats.bytes != old_stats.bytes) {
		n = 0;
		sprintf(buf, "Cum bytes: %ld", stats.bytes);
		XtSetArg(args[n], XtNlabel, buf); n++;
		XtSetValues(w_status[1], args, n);
		}

	n = 0;
	t1 = t - stats.start_time;
	sprintf(buf, "Total/sec: %ld", stats.tot_bytes / (t1 ? t1 : 1));
	XtSetArg(args[n], XtNlabel, buf); n++;
	XtSetValues(w_status[2], args, n);

	n = 0;
	t1 = t - stats.time;
	sprintf(buf, "Cum/sec: %ld", stats.tot_bytes / (t1 ? t1 : 1));
	XtSetArg(args[n], XtNlabel, buf); n++;
	XtSetValues(w_status[3], args, n);
	/***********************************************/
	/*   Start timer again.			       */
	/***********************************************/
restart_timer:
	timer_id = XtAppAddTimeOut(app_con, 1000L, timeout_proc, NULL);

	old_stats = stats;
}
/**********************************************************************/
/*   Function called when CANCEL button selected in font dialog box.  */
/**********************************************************************/
void
cancel_font(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{
	XtDestroyWidget(font_dialog);
	font_dialog = (Widget) NULL;
}
/**********************************************************************/
/*   Function called when OK button selected in font dialog box.      */
/**********************************************************************/
void
ok_font(w, client_data, call_data)
Widget	w;
XtPointer	client_data, call_data;
{	char	*font;
	int	n;
	Arg	args[20];
	int	rows, columns;
	int	fwidth, fheight;


	font = XawToggleGetCurrent(w_fonts[0]);
	XtDestroyWidget(font_dialog);
	font_dialog = (Widget) NULL;

	if (font) {
		n = 0;
		XtSetArg(args[n], XtNfont, font); n++;
		XtSetValues(ctw_widget, args, n);

		get_font_size(ctw_widget, &fwidth, &fheight);

		n = 0;
		XtSetArg(args[n], XtNrows, &rows); n++;
		XtSetArg(args[n], XtNcolumns, &columns); n++;
		XtGetValues(ctw_widget, args, n);
		/***********************************************/
		/*   Re-use the row column size.	       */
		/***********************************************/
		n = 0;
		XtSetArg(args[n], XtNrows, rows); n++;
		XtSetArg(args[n], XtNcolumns, columns); n++;
		XtSetArg(args[n], XtNwidth, fwidth * columns); n++;
		XtSetArg(args[n], XtNheight, fheight * rows); n++;
		XtSetValues(ctw_widget, args, n);

		n = 0;
		XtSetArg(args[n], XtNheight, fheight * rows); n++;
		XtSetValues(pane, args, n);
		XtSetValues(scroll_bar, args, n);
		}
}
