/**********************************************************************/
/*   Color  terminal  emulator using the ctw (Color terminal widget)  */
/*   widget.							      */
/*   								      */
/*   This  file  contains  routines  which  are common to the athena  */
/*   and motif versions of fcterm.				      */
/*   								      */
/*   Author: Paul Fox						      */
/*   Date:   February 1992       				      */
/**********************************************************************/
# include	"fcterm.h"

int	console_flag = FALSE;
char	*geometry;
int	geom_x = -1, geom_y = -1;
int	geom_width = -1, geom_height = -1;
char	*term_str = "TERM=xterm";
char	*bterm_str = "BTERM=xcrisp-col";
char	*foxterm_str = "FOXTERM=xterm";

Widget		top_level;
Widget		ctw_widget;

struct stats  stats, old_stats;
XtIntervalId	timer_id;

XtAppContext	app_con;
int	pty_fd, tty_fd;
int	pid;

enum emulation emulator = EM_NONE;

/**********************************************************************/
/*   Name  of  main  widget  can  be changed depending on whether we  */
/*   are  in  ISC  or  SCO  mode  so  that user can set up different  */
/*   Xdefaults.							      */
/**********************************************************************/
char	*widget_name = "pane";
/**********************************************************************/
/*   Copy of attributes which user can change.			      */
/**********************************************************************/
char	**attr_names;

/**********************************************************************/
/*   Trap certain types of signals.				      */
/**********************************************************************/
void
setup_signal_handlers()
{
	signal(SIGCHLD, child_handler);
	signal(SIGINT, int_handler);
	signal(SIGTERM, int_handler);
	signal(SIGHUP, int_handler);
}
/**********************************************************************/
/*   Grab  switches  before  X  sees  them. We need to grab geometry  */
/*   before  X  sees  them because we want geometry to be treated in  */
/*   width*cols  and  we want it to apply to the VT100 widget -- not  */
/*   necessarily the application shell.				      */
/**********************************************************************/
int
steal_switches(argc, argv)
int	argc;
char	**argv;
{	int	i, j;
	char	*cp;

	prog_name = argv[0];
	for (i = 1; i < argc; ) {
		if (strcmp(argv[i], "-isc") == 0) {
			term_str = "TERM=AT386";
			foxterm_str = "FOXTERM=isc";
			bterm_str = "BTERM=isc";
			emulator = EM_ISC;
			widget_name = "iscterm";
			}
		else if (strcmp(argv[i], "-sco") == 0) {
			term_str = "TERM=AT386";
			foxterm_str = "FOXTERM=sco";
			bterm_str = "BTERM=sco";
			emulator = EM_SCO;
			widget_name = "scoterm";
			}
		else if (strncmp(argv[i], "-geo", 4) == 0) {
			/***********************************************/
			/*   Leave  the  (x,y)  co-ords  intact. Just  */
			/*   steal the window size.		       */
			/***********************************************/
			if (argv[++i][0] >= '0' && argv[i][0] <= '9') {
				geometry = argv[i];
				for (cp = argv[i]; *cp && *cp != '-' && *cp != '+'; )
					cp++;
				if (*cp == '\0') {
					/***********************************************/
					/*   If  we  only  have WxH then delete whole  */
					/*   argument.				       */
					/***********************************************/
					for (j = --i; j < argc; j++) {
						argv[j] = argv[j+1];
						}
					argc--;
					}
				else
					argv[i] = cp;
				}
			i++;
			continue;
			}
		else {
			i++;
			continue;
			}
		for (j = i; j < argc; j++) {
			argv[j] = argv[j+1];
			}
		argc--;
		}
	return argc;
	
}
# include	"fcterm.bitmap"
void
create_icon_pixmap()
{	Pixmap	pixmap;
	Arg	args[20];

	pixmap = XCreateBitmapFromData(XtDisplay(top_level), 
		XtWindow(top_level),
		fcterm_bits, fcterm_width, fcterm_height);
	XtSetArg(args[0], XtNiconPixmap, pixmap);
	XtSetValues(top_level, args, 1);
}
/**********************************************************************/
/*   Routine  to  set the default attributes of the widget depending  */
/*   on whether user want SCO/ISC functionality.		      */
/**********************************************************************/
void
set_default_attributes()
{	int	len;
	int	*flags;
	int	attributes[64];

	/***********************************************/
	/*   Set   attributes  depending  on  command  */
	/*   line terminal type.		       */
	/***********************************************/
	len = ctw_get_attributes(ctw_widget, &flags, &attr_names);
	memcpy((char *) attributes, (char *) flags, len * sizeof (int));
	switch (emulator) {
	  case EM_ISC:
	  	attributes[CTW_ISC_DRIVER] = TRUE;
	  	attributes[CTW_ESC_LITERAL] = TRUE;
		break;
	  case EM_SCO:
	  	attributes[CTW_SCO_DRIVER] = TRUE;
		break;
	  default:
	  	break;
	  }
	ctw_set_attributes(ctw_widget, attributes);
}
/**********************************************************************/
/*   Function to handle keyboard input. Just send it to the pty.      */
/**********************************************************************/
void
input_callback(w, type, reason)
Widget	w;
int	type;
ctw_callback_t	*reason;
{
	if (reason->len)
		write(pty_fd, reason->ptr, reason->len);
}
/**********************************************************************/
/*   Function to handle application oriented escape sequences.	      */
/**********************************************************************/
void
appl_callback(w, type, reason)
Widget	w;
int	type;
ctw_callback_t	*reason;
{	int	fht, fwd;

	switch (reason->reason) {
	  case CTW_OPEN_WINDOW:
	  	XtVaSetValues(top_level, XtNiconic, FALSE, NULL);
	  	break;
	  case CTW_CLOSE_WINDOW:
	  	XtVaSetValues(top_level, XtNiconic, TRUE, NULL);
	  	break;
	  case CTW_MOVE_WINDOW:
	  	XtVaSetValues(top_level,
			XtNx,	reason->x,
			XtNy,	reason->y,
			NULL);
		break;
	  case CTW_SIZE_WINDOW_PIXELS:
	  	XtVaSetValues(top_level,
			XtNwidth,	reason->width,
			XtNheight,	reason->height,
			NULL);
		break;
	  case CTW_FRONT_WINDOW:
	  	XRaiseWindow(XtDisplay(top_level), XtWindow(top_level));
		break;
	  case CTW_BACK_WINDOW:
	  	XLowerWindow(XtDisplay(top_level), XtWindow(top_level));
		break;
	  case CTW_SIZE_WINDOW_CHARS:
	  	get_font_size(w, &fwd, &fht);
	  	XtVaSetValues(top_level,
			XtNwidth,	fwd * reason->width,
			XtNheight,	fht * reason->height,
			NULL);
		break;
	  default:
	  	break;
	  }
}
/**********************************************************************/
/*   Function to handle input from pty.				      */
/**********************************************************************/
void
pty_input(client_data, src, id)
XtPointer	client_data;
int		*src;
XtInputId	*id;
{
# define	BUF_SIZE	4096
	char	buf[BUF_SIZE + 1];
	int	n;

	n = read(pty_fd, buf, BUF_SIZE);
	if (n < 0) {
		if (errno != EIO)
			perror("read");
		exit(1);
		}
	if (n == 0) {
		fprintf(stderr, "EOF detected on pty\n");
		exit(1);
		}
	buf[n] = 0;
	ctw_add_string(ctw_widget, buf, n);
	stats.bytes += n;
	stats.tot_bytes += n;
/*	XFlush(XtDisplay(top_level));*/
}
/**********************************************************************/
/*   Function to create a child shell process.			      */
/**********************************************************************/
void
spawn_child()
{
	int	i, j;
	char	buf[BUFSIZ];
	char	*getenv();
	int	one = 1;
	
	for (i = 'p'; i < 'z'; i++) {
		for (j = 0; j < 16; j++) {
			sprintf(buf, "/dev/pty%c%x", i, j);
			if ((pty_fd = open(buf, O_RDWR)) >= 0)
				goto found_pty;
			}
		}
	fprintf(stderr, "Couldn't find a free pty\n");
	exit(1);
found_pty:
	switch (pid = vfork()) {
	  case -1:
	  	perror("fork");
		exit(1);
	  case 0: {
	  	struct termio t;
		char	*shell = getenv("SHELL");
		if (shell == NULL)
			shell = "/bin/sh";

		putenv(term_str);
		putenv(foxterm_str);
		putenv(bterm_str);

	  	close(0);
	  	close(1);
	  	close(2);
		setpgrp(getpid(), 0);
		sprintf(buf, "/dev/tty%c%x", i, j);
		tty_fd = open(buf, O_RDWR);
		if (tty_fd < 0) {
			perror(buf);
			exit(1);
			}
		dup(tty_fd);
		dup(tty_fd);
		
		ioctl(0, TCGETA, &t);
		t.c_cflag &= ~PARENB;
		t.c_cflag |= CS8;
		t.c_oflag &= ~TAB3;
		t.c_iflag |= IGNPAR;
		t.c_lflag |= ECHOCTL | ECHOKE | ECHOE | ECHOK;
		ioctl(0, TCSETA, &t);

# if defined(TIOCCONS)
		if (console_flag)
			ioctl(0, TIOCCONS, (char *) &one);
# endif

		signal(SIGTERM, SIG_DFL);

		execlp(shell, shell, "-i", (char *) 0);
		_exit(1);
		}
	  default:
	  	break;
	  }
}
void
child_handler()
{
	int	status, ret;

	ret = wait(&status);
	if (status)
		printf("Child process died exit code: %d\n", status);
	exit(1);
}
/**********************************************************************/
/*   Function to handle notification of window being resized.	      */
/**********************************************************************/
void
resize_callback(w, type, reason)
Widget	w;
int	type;
ctw_callback_t	*reason;
{
	set_window_size(w);
}
/**********************************************************************/
/*   Tell PTY that the window size has changed.			      */
/**********************************************************************/
void
set_window_size(w)
CtwWidget	w;
{	int	n;
	Arg	args[20];
	static struct winsize	winsize;
	int	rows, columns;
	
	n = 0;
	XtSetArg(args[n], XtNrows, &rows); n++;
	XtSetArg(args[n], XtNcolumns, &columns); n++;
	XtGetValues(w, args, n);

	if (winsize.ws_row == rows && winsize.ws_col == columns)
		return;
		
	winsize.ws_row = rows;
	winsize.ws_col = columns;
	ioctl(pty_fd, TIOCSWINSZ, &winsize);
}
void
int_handler()
{
	if (pid) {
		kill(pid, SIGTERM);
		kill(pid, SIGHUP);
		kill(pid, SIGKILL);
		}
	exit(2);
}
int
do_switches(argc, argv)
int	argc;
char	**argv;
{	int	i;
	char	*cp;

	for (i = 1; i < argc; i++) {
		cp = argv[i];
		if (*cp++ != '-')
			return i;
		if (strcmp(cp, "ut") == 0)
			continue;
		if (strncmp(cp, "geom", 4) == 0) {
			if (++i >= argc)
				usage();
			cp = argv[i];
			if (*cp == '+')
				sscanf(cp, "+%d+%d", &geom_x, &geom_y);
			else
				sscanf(cp, "%dx%d+%d+%d", 
					&geom_width, &geom_height,
					&geom_x, &geom_y);
			continue;
			}
		while (*cp) {
			switch (*cp++) {
			  case 'C':
			  	console_flag = TRUE;
				break;
			  default:
			  	usage();
			  }
			}
		}
	return i;
}
void
usage()
{
	fprintf(stderr, "%s [-isc | -sco] [standard X11 options] [-ut]\n", prog_name);
	fprintf(stderr, "	[-geometry WxH+X+Y] [-C]\n");
	fprintf(stderr, "	-isc	Emulate Interactive Unix console driver.\n");
	fprintf(stderr, "	-sco	Emulate SCO Unix console driver.\n");
	fprintf(stderr, "	-ut	Inhibit utmp entry.\n");
	fprintf(stderr, "		(Currently unimplemented)\n");
	exit(1);
}
/**********************************************************************/
/*   Get the current font size for the Ctw widget.		      */
/**********************************************************************/
void
get_font_size(w, fwidth, fheight)
CtwWidget	w;
int	*fwidth;
int	*fheight;
{
	int	len;
	int	*flags;

	len = ctw_get_attributes(ctw_widget, &flags, &attr_names);
	*fwidth = flags[CTW_FONT_SIZE] >> 16;
	*fheight = flags[CTW_FONT_SIZE] & 0xffff;
}

